@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 import argparse
 import copy
 import getpass
@@ -20,6 +20,7 @@ from typing import Callable, Dict, List, Optional, Set, Tuple
 from urllib.parse import urlsplit
 
 import termcolor
+import timeout_decorator
 import yaml
 from ansi2html import Ansi2HTMLConverter
 from dateutil import parser
@@ -107,8 +108,7 @@ def dt_challenges_evaluator_(args: List[str] = None, token: str = None):
     from duckietown_challenges.col_logging import setup_logging
 
     setup_logging()
-    elogger.info("dt-challenges-evaluator (DTC %s)" % __version__)
-    elogger.info("called with:\n%s" % sys.argv)
+    elogger.info(f"dt-challenges-evaluator {__version__}", args=args, sys_argv=sys.argv)
 
     usage = """
 
@@ -143,10 +143,10 @@ Usage:
         msg = "Cannot specify both --submission and --continuous."
         raise UserError(msg)
     if parsed.version:
-        msg = "Duckietown Challenges Runner %s" % __version__
+        msg = f"Duckietown Challenges Runner {__version__}"
         from duckietown_challenges import __version__ as dcversion
 
-        msg += "\nDuckietown Challenges %s" % dcversion
+        msg += f"\nDuckietown Challenges {dcversion}"
         print(msg)
         return
 
@@ -184,24 +184,6 @@ Usage:
     if token is None:
         token = get_token_from_shell_config()
 
-    args = dict(
-        do_upload=do_upload,
-        do_pull=do_pull,
-        more_features=more_features,
-        delete=delete,
-        evaluator_name=evaluator_name,
-        machine_id=machine_id,
-        tmpdir=tmpdir,
-        token=token,
-        debug_volumes=parsed.debug_volumes,
-        impersonate=parsed.impersonate,
-        copy_to_machine_cache=copy_to_machine_cache,
-        allow_host_network=allow_host_network,
-        use_ipfs=parsed.ipfs,
-        do_not_mess_with_repo=parsed.do_not_mess_with_repo,
-    )
-    ndots = 0
-
     if parsed.continuous:
 
         timeout = 5.0  # seconds
@@ -212,19 +194,37 @@ Usage:
             multiplier = min(multiplier, max_multiplier)
             t0 = time.time()
             try:
-                go_(None, reset=False, **args)
+                go_(
+                    None,
+                    reset=False,
+                    do_upload=do_upload,
+                    do_pull=do_pull,
+                    more_features=more_features,
+                    delete=delete,
+                    evaluator_name=evaluator_name,
+                    machine_id=machine_id,
+                    tmpdir=tmpdir,
+                    token=token,
+                    debug_volumes=parsed.debug_volumes,
+                    impersonate=parsed.impersonate,
+                    copy_to_machine_cache=copy_to_machine_cache,
+                    allow_host_network=allow_host_network,
+                    use_ipfs=parsed.ipfs,
+                    do_not_mess_with_repo=parsed.do_not_mess_with_repo,
+                )
                 multiplier = 1.0
                 if parsed.one:
                     msg = "Because --one was passed, I will finish here."
+                    elogger.info(msg)
                     break
             except NothingLeft:
                 delta = time.time() - t0
 
-                sys.stderr.write("No work for me yet. Server answers in %.1f seconds.\n" % delta)
-                ndots += 1
-                if ndots == 5:
-                    # sys.stderr.write(' no work for me yet.\n')
-                    ndots = 0
+                elogger.debug(f"No work for me yet. Server answers in {delta:.1f} seconds.")
+                # ndots += 1
+                # if ndots == 5:
+                #     # sys.stderr.write(' no work for me yet.\n')
+                #     ndots = 0
                 # time.sleep(timeout * multiplier)
                 # elogger.info('No submissions available to evaluate.')
             except ConnectionError as e:
@@ -233,8 +233,8 @@ Usage:
             except KeyboardInterrupt:
                 break
             except BaseException:
-                msg = "Uncaught exception:\n%s" % traceback.format_exc()
-                elogger.error(msg)
+                msg = "Uncaught exception"
+                elogger.error(msg, e=traceback.format_exc())
                 multiplier *= multiplier_grow
 
             r = random.random() * 2
@@ -252,18 +252,35 @@ Usage:
 
         for submission_id in submissions:
             try:
-                go_(submission_id, reset=reset, **args)
+                go_(
+                    submission_id,
+                    reset=reset,
+                    do_upload=do_upload,
+                    do_pull=do_pull,
+                    more_features=more_features,
+                    delete=delete,
+                    evaluator_name=evaluator_name,
+                    machine_id=machine_id,
+                    tmpdir=tmpdir,
+                    token=token,
+                    debug_volumes=parsed.debug_volumes,
+                    impersonate=parsed.impersonate,
+                    copy_to_machine_cache=copy_to_machine_cache,
+                    allow_host_network=allow_host_network,
+                    use_ipfs=parsed.ipfs,
+                    do_not_mess_with_repo=parsed.do_not_mess_with_repo,
+                )
             except NothingLeft as e:
                 if submission_id is None:
                     msg = "No submissions available to evaluate."
                 else:
-                    msg = "Could not evaluate submission %s." % submission_id
+                    msg = f"Could not evaluate submission {submission_id}."
 
                 msg += "\n" + str(e)
                 elogger.error(msg)
 
 
-class NothingLeft(Exception):
+class NothingLeft(ZException):
     pass
 
 
@@ -313,7 +330,7 @@ def get_features(more_features, use_ipfs: bool):
 
     for k, v in more_features.items():
         if k in features:
-            msg = "Using %r = %r instead of %r" % (k, more_features[k], features[k])
+            msg = f"Using {k!r} = {more_features[k]!r} instead of {features[k]!r}"
             # elogger.info(msg)
         features[k] = v
 
@@ -453,7 +470,7 @@ def go_(
 
     elogger.info(f"Using server_host", server_host=server_host, url=url)
 
-    evaluator_version = "d-c:%s;d-c-r:%s" % (version_dc, version_dcr,)
+    evaluator_version = f"d-c:{version_dc};d-c-r:{version_dcr}"
     process_id = evaluator_name
 
     timeout_server = 60
@@ -471,8 +488,8 @@ def go_(
 
     if "job_id" not in res:
         elogger.info("No jobs available", url=url)
-        msg = "Could not find jobs: %s" % res["msg"]
-        raise NothingLeft(msg)
+        msg = "Could not find jobs."
+        raise NothingLeft(msg, res=res)
     job_id = res["job_id"]
     elogger.info(res=res)
 
@@ -552,10 +569,10 @@ def go_(
                 timeout=timeout_server,
             )
             break
-        except BaseException as e:
-            msg = "Could not report: %s" % e
-            elogger.warning(msg)
-            elogger.info("Retrying %s more times after %s seconds" % (ntries, interval))
+        except BaseException:
+            msg = "Could not report."
+            elogger.warning(msg, e=traceback.format_exc())
+            elogger.info(f"Retrying {ntries} more times after {interval} seconds")
             ntries -= 1
             time.sleep(interval)
 
@@ -576,6 +593,7 @@ def get_cr(
     cd: Optional[ChallengeDescription],
     heartbeat: Optional[Callable] = None,
 ) -> Tuple[ChallengeResults, List]:
+
     # noinspection PyBroadException
     try:
 
@@ -584,12 +602,12 @@ def get_cr(
         submission_id = res["submission_id"]
         timeout_sec = res["timeout"]
         if timeout_sec is None or timeout_sec == 0:
-            msg = "Invalid timeout %s" % timeout_sec
+            msg = f"Invalid timeout {timeout_sec}"
             raise ValueError(msg)
 
     except BaseException:
-        msg = "Uncaught exception:\n%s" % traceback.format_exc()
-        elogger.error(msg)
+        msg = "Uncaught exception."
+        elogger.error(msg, e=traceback.format_exc())
         status = ChallengesConstants.STATUS_JOB_HOST_ERROR
         return ChallengeResults(status, msg, scores={}), []
 
@@ -605,21 +623,23 @@ def get_cr(
         organization = location["organization"]
         repository = location["repository"]
         tag = location["tag"]
-        digest = location["digest"]
+        digest = location["digest"]  # FIXME: previously disabled, now we should re-enable
         registry = location["registry"]
 
         if registry == "dockerhub":
-            solution_container = "%s/%s:%s" % (organization, repository, tag)
-        # solution_container = '%s/%s:%s@%s' % (organization, repository, tag, digest)
-        else:
-            solution_container = "%s/%s/%s:%s" % (registry, organization, repository, tag,)
-        elogger.info("solution_container = %s" % solution_container)
+            registry = "docker.io"
+
+            # solution_container = f"{organization}/{repository}:{tag}"
+        solution_container = f"{registry}/{organization}/{repository}:{tag}@{digest}"
+        # else:
+        #     solution_container = f"{registry}/{organization}/{repository}:{tag}"
+        elogger.info(f"solution_container = {solution_container}")
 
         wd = os.path.join(
             tmpdir,
             challenge_name,
-            "submission%d" % submission_id,
-            "%s-%s-job%s" % (challenge_step_name, evaluator_name, job_id),
+            f"submission{submission_id}",
+            f"{challenge_step_name}-{evaluator_name}-job{job_id}",
         )
 
         if not os.path.exists(wd):
@@ -640,6 +660,7 @@ def get_cr(
         try:
 
             with setup_logging(own_logs_dir):
+                elogger.debug("running get_cr()", delete=delete, copy_to_machine_cache=copy_to_machine_cache)
                 challenge_parameters_ = EvaluationParameters.from_yaml(res["challenge_parameters"])
 
                 rnd = random.randint(1, 1000000)
@@ -670,28 +691,20 @@ def get_cr(
             if do_upload:
                 uploaded = upload_files(wd, aws_config, copy_to_machine_cache=copy_to_machine_cache)
 
-        if delete:
-            try:
-                cmd = ["down"]
-                run_docker(wd, project, cmd)
-            except:
-                elogger.warning("While taking down after success:\n\n" + traceback.format_exc())
-
-        if delete:
-            if os.path.exists(wd):
-                shutil.rmtree(wd)
-        else:
-            msg = f"I will not delete temporary dir {wd}"
-            elogger.info(msg)
+        try:
+            cmd = ["down"]
+            run_docker(wd, project, cmd)
+        except DockerComposeFail:
+            elogger.warning("While taking down after success:\n\n" + traceback.format_exc())
 
     except KeyboardInterrupt:
-        msg = "KeyboardInterrupt:\n%s" % traceback.format_exc()
+        msg = f"KeyboardInterrupt:\n{traceback.format_exc()}"
 
         cmd = ["down"]
         if wd is not None:
             try:
                 run_docker(wd, project, cmd)
-            except:
+            except DockerComposeFail:
                 elogger.warning("While taking down after failure:\n\n" + traceback.format_exc())
 
         elogger.error(msg)
@@ -699,16 +712,28 @@ def get_cr(
         cr = ChallengeResults(status, msg, scores={})
 
     except IPFSException:
-        msg = "Could not access IPFS data:\n%s" % traceback.format_exc()
+        msg = f"Could not access IPFS data:\n{traceback.format_exc()}"
         elogger.error(msg)
         status = ChallengesConstants.STATUS_JOB_HOST_ERROR
         cr = ChallengeResults(status, msg, scores={})
 
     except BaseException:
-        msg = "Uncaught exception:\n%s" % traceback.format_exc()
+        msg = f"Uncaught exception:\n{traceback.format_exc()}"
         elogger.error(msg)
         status = ChallengesConstants.STATUS_JOB_HOST_ERROR
         cr = ChallengeResults(status, msg, scores={})
+    finally:
+        if wd:
+            if delete:
+                msg = f"I am deleting temporary dir {wd}"
+                elogger.info(msg)
+                if os.path.exists(wd):
+                    shutil.rmtree(wd)
+            else:
+                msg = f"I will not delete temporary dir {wd}"
+                elogger.info(msg)
+        else:
+            elogger.info("Temp dir wd not set")
 
     return cr, uploaded
 
@@ -876,7 +901,7 @@ def write_logs_worker(service_name, container_id):
         timestamp = x[:i]
         message = x[i + 1 :].rstrip()  # \n
         ts = parser.parse(timestamp)
-        s = "%s %s %s" % (ts.strftime("%H:%M:%S"), service_name, message)
+        s = f"{ts.strftime('%H:%M:%S')} {service_name} {message}"
         sys.stdout.write(s + "\n")
         sys.stdout.flush()
         if container.status != "running":
@@ -884,7 +909,7 @@ def write_logs_worker(service_name, container_id):
 
     exit_code = exit_code_for_container(container_id)
 
-    elogger.info("container %s terminated with %s" % (service_name, exit_code))
+    elogger.info(f"container {service_name} terminated with {exit_code}")
 
 
 def exit_code_for_container(container_id):
@@ -1018,7 +1043,7 @@ def run_one(
                     exited.append(service_name)
                     exit_code = exit_code_for_container(container_id)
                     if exit_code:
-                        msg = 'The container "%s" exited with code %s.\n' % (service_name, exit_code,)
+                        msg = f'The container "{service_name}" exited with code {exit_code}.\n'
                         if exit_code == 137:
                             msg += (
                                 "\n\nError code 137 likely means an out-of-memory error. You likely did not "
@@ -1038,11 +1063,9 @@ def run_one(
                     running.append(service_name)
 
             delta = time.time() - t0
-            msg = "Running for %d s (timeout: %d s). Services running: %s  exited:  %s" % (
-                delta,
-                timeout_sec,
-                running,
-                exited,
+            msg = (
+                f"Running for {int(delta)} s (timeout: {int(timeout_sec)} s). Services running: {running}  "
+                f"exited:  {exited}"
             )
             elogger.debug(msg)
 
@@ -1076,19 +1099,19 @@ Check the evaluator log to see what happened."""
             cr = ChallengeResults(status, msg, scores={})
 
     except DockerComposeFail as e:
-        msg = "Error while running Docker Compose:\n\n%s" % e
+        msg = f"Error while running Docker Compose:\n\n{e}"
         elogger.error(msg)
         status = ChallengesConstants.STATUS_JOB_HOST_ERROR
         cr = ChallengeResults(status, msg, scores={})
 
     except KeyboardInterrupt:
-        msg = "KeyboardInterrupt:\n%s" % traceback.format_exc()
+        msg = f"KeyboardInterrupt:\n{traceback.format_exc()}"
         elogger.error(msg)
         status = ChallengesConstants.STATUS_JOB_ABORTED
         cr = ChallengeResults(status, msg, scores={})
 
     except BaseException:
-        msg = "Uncaught exception while running Docker Compose:\n%s" % traceback.format_exc()
+        msg = f"Uncaught exception while running Docker Compose:\n{traceback.format_exc()}"
         elogger.error(msg)
         status = ChallengesConstants.STATUS_JOB_HOST_ERROR
         cr = ChallengeResults(status, msg, scores={})
@@ -1356,9 +1379,11 @@ class Tee:
 
 
 def run_docker(cwd: str, project: str, cmd0: List[str], get_output: bool = False) -> bytes:
+
+    """ raises DockerComposeFail """
     cmd0 = ["docker-compose", "-p", project] + cmd0
     # elogger.info('Running:\n\t%s' % " ".join(cmd0) + '\n\n in %s' % cwd)
-    elogger.debug("Running: %s" % " ".join(cmd0))
+    elogger.debug(command=" ".join(cmd0))
 
     d = tempfile.mkdtemp()
 
@@ -1377,10 +1402,10 @@ def run_docker(cwd: str, project: str, cmd0: List[str], get_output: bool = False
             # subprocess.check_call(cmd0, cwd=cwd, stdout=tee_stdout, stderr=tee_stderr)
             subprocess.check_call(cmd0, cwd=cwd, stdout=sys.stdout, stderr=sys.stderr)
     except subprocess.CalledProcessError as e:
-        msg = "Could not run %s:\n\n %s" % (cmd0, indent(e, "  >  "))
-        msg += "\n\n%s" % indent(tee_stdout.getvalue(), "stdout | ")
-        msg += "\n\n%s" % indent(tee_stderr.getvalue(), "stderr | ")
-        raise DockerComposeFail(msg) from e
+        msg = "Could not run command"
+        raise DockerComposeFail(
+            msg, cmd=cmd0, stdout=tee_stdout.getvalue(), sderr=tee_stderr.getvalue(), e=str(e)
+        ) from e
     finally:
         try:
             os.unlink(fn1)
@@ -1408,9 +1433,6 @@ def upload_files(
     return uploaded
 
 
-import timeout_decorator
-
-
 @timeout_decorator.timeout(60)
 def try_to_get_hash(client, qm):
     res = client.file_ls(qm)
